
/*
 * Copyright (c) 1999 Horace Ho
 * All rights reserved.
 *
 * MemoEnDe.c
 *
 */


#include <Pilot.h>
#include <crc.h>
#include <Encrypt.h>
#include "MemoRsc.h"
#include "MemoEnDE.h"

PrefType pref;
EnDeType ende;

#define UD_NO_ERROR			0
#define UD_ENDE_DES			sysFtrNumEncryptionMaskDES

#define UD_ENDE_SIGNATURE	"\nEncoded memo - do not edit !!\n  [xxxx] "
#define UD_ENDE_CHECKSUM	"ffff \n\n"

#define noRecordSelected	0xffff

#define UD_WAIT_BITMAP_X	28
#define UD_WAIT_BITMAP_Y	148

// #define UD_BLINK_WAIT_BITMAP

static Char HEXDEC[16] = "0123456789ABCDEF";
static Char BASE64[64] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
static Byte ASCII[128] = {	255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,  //  15
							255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,  //  31
							255, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255,  62, 255, 255, 255,  63,  //  47
							 52,  53,  54,  55,  56,  57,  58,  59,  60,  61, 255, 255, 255,   0, 255, 255,  //  63
							255,   0,   1,   2,   3,   4,   5,   6,  7,    8,   9,  10,  11,  12,  13,  14,  //  79
							 15,  16,  17,  18,  19,  20,  21,  22,  23,  24,  25, 255, 255, 255, 255, 255,  //  95
							255,  26,  27,  28,  29,  30,  31,  32,  33,  34,  35,  36,  37,  38,  39,  40,  // 111
							 41,  42,  43,  44,  45,  46,  47,  48,  49,  50,  51, 255, 255, 255, 255, 255 };// 127

// MemoCreateRecord - done
// MemoSaveRecord   -
// MemoDeleteRecord - done

Boolean ende_init()
{
	Boolean ok = true;
	Err rc = FtrGet(sysFtrCreator, sysFtrNumEncryption, &ende.encrypt_list);
	if (rc) ende.encrypt_list = 0;
 //	ende.encrypt_list = UD_ENDE_DES;
	ende.encrypt_user = UD_ENDE_DES;
	ende.record = noRecordSelected;
	ende.encode = false;
	ende_set_signature(0);
 //	ende_set_password(NULL);
	StrCopy(ende.checksum, UD_ENDE_CHECKSUM);
 //	ende_set_checksum(ende.checksum, 'EnDe');
	
	if (ende.encrypt_list == UD_NO_ENDE) {
		FrmAlert(NoEndeAlert);
	} else {
		if (!ende_digest_defed(pref.pwdod, UD_DIGEST_SIZE) && !ende_password_1st()) {
			ok = false;
		} else {
			Word x = 0x100 * pref.pwdod[0] + pref.pwdod[1];
			ende_set_signature(x);
		}
	}
	return ok;
}

void ende_set_signature(Word x)
{
	StrNCopy(ende.signature, UD_ENDE_SIGNATURE, sizeof(ende.signature));
	ende.signature[sizeof(ende.signature)-1] = '\0';
	ende.sign_size = StrLen(ende.signature);
	ende.csum_size = sizeof(ende.checksum)-1;
	ende.head_size = ende.sign_size + ende.csum_size;
	{
	char* xxxx = StrStr(ende.signature, "xxxx");
	if (xxxx) ende_set_checksum(xxxx, x);
	}
}

void ende_set_checksum(char *s, Word x)
{
	s[0] = HEXDEC[(x/0x1000) & 0x0f];
	s[1] = HEXDEC[(x/0x100)  & 0x0f];
	s[2] = HEXDEC[(x/0x10)   & 0x0f];
	s[3] = HEXDEC[(x)        & 0x0f];
}

Word ende_get_checksum(char *s)
{
	Word i; Word cs = 0;
	for (i = 0; i < 4; i++) {
		if (s[i] >= '0' && s[i] <= '9') {
			cs += s[i] - '0';
		} else if (s[i] >= 'A' && s[i] <= 'F') {
			cs += s[i] - 'A' + 0x0a;
		}
		if (i < 3) cs *= 0x10;
	}
	return cs;
}

Boolean ende_is_record_encoded(DmOpenRef db, Word record)
{
	Boolean  is_encoded = false;

	VoidHand hand = DmQueryRecord(db, record);
	if (hand) {
		char* ptr = MemHandleLock(hand);
		if (ptr) {
			// look for EnDe signatur
			ULong title_size;
			ULong coded_size;
			Err rc;
			is_encoded = ende_is_text_encoded(ptr, &title_size, &coded_size);
			rc = MemHandleUnlock(hand);
		}
	}
	return is_encoded;
}

Boolean ende_is_text_encoded(const char* s, ULong *title_size, ULong *coded_size)
{
	Boolean is_encoded = false;
	Word  checksum = 0;
	char* head_ptr;
	ULong rest_len = 0;

	*title_size = 0;
	*coded_size = 0;

	head_ptr = StrStr(s, ende.signature);
	if (head_ptr) {
		*title_size = head_ptr-s;
		rest_len = StrLen(head_ptr);
	}

	if (rest_len > ende.head_size) {
		*coded_size = rest_len - ende.head_size;
		checksum = Crc16CalcBlock(head_ptr+ende.head_size, *coded_size, 0);
		if (checksum == ende_get_checksum(head_ptr+ende.sign_size)) {
			is_encoded = true;
		}
	}

	return (is_encoded);
}

Boolean ende_get_record_encode(Word record)
{
	Boolean encode = false;
	if (record != noRecordSelected && record == ende.record) {
		encode = ende.encode;
	}
	ende.record = record;
	return encode;
}

void ende_set_record_encode(Word record, Boolean encode)
{
	ende.record = record;
	ende.encode = encode;
}

void ende_decode_record(DmOpenRef db, Word record)
{
	Boolean  is_encoded = false;
	VoidHand old_hand = NULL;
	VoidHand new_hand = NULL;
	ULong    old_size;			// original record size
	ULong    new_size;			// encoded record size
	ULong    title_size = 0;	// size of title (begin to 1st '\n' or eof)
	ULong    coded_size = 0;	// size of encoded text
	char*    old_ptr;
	char*    new_ptr;
	Err      rc;

	old_hand = DmQueryRecord(db, record);
	if (old_hand) {
		old_size = MemHandleSize(old_hand);
		old_ptr = MemHandleLock(old_hand);
		if (old_ptr) {
			// look for EnDe signatur
			if (ende_is_text_encoded(old_ptr, &title_size, &coded_size)) {
				uiDrawBitmap(EditWaitBitmap, UD_WAIT_BITMAP_X, UD_WAIT_BITMAP_Y);
				new_hand = ende_decode_text(old_ptr, title_size, ende.head_size, coded_size);
			 	uiEraseRect(UD_WAIT_BITMAP_X, UD_WAIT_BITMAP_Y, 10, 10);
			}
			rc = MemHandleUnlock(old_hand);
		}
	}

	if (new_hand) {
		new_size = MemHandleSize(new_hand);
		new_ptr = MemHandleLock(new_hand);
		old_hand = DmResizeRecord(db, record, new_size);
		if (old_hand) {
			old_ptr = MemHandleLock(old_hand);
		} else {
			old_ptr = NULL;
		}
		if (old_ptr && new_ptr) {
			rc = DmWrite(old_ptr, 0, new_ptr, new_size);
			MemSet(new_ptr, new_size, 0);
			rc = MemHandleUnlock(old_hand);
			rc = MemHandleUnlock(new_hand);
		} else {
			if (old_ptr) rc = MemHandleUnlock(old_hand);
			if (new_ptr) rc = MemHandleUnlock(new_hand);
		}
		if (old_hand) {
			rc = DmReleaseRecord(db, record, true);
		}
		rc = MemHandleFree(new_hand);

		is_encoded = true;
	} else {
		is_encoded = false;
	}

	ende_set_record_encode(record, is_encoded);

}

VoidHand ende_decode_text(char *rec, ULong title_size, ULong header_size, ULong coded_size)
{
	VoidHand h_crypt = NULL;
	VoidHand h_total = NULL;
	ULong crypt_size, plain_size, total_size;
	Err rc;
	Boolean ok = false;

	if (ende_password_ok(ende.password, UD_DIGEST_SIZE)) {
		ok = true;
	} else {
		ende_password_set(ende.password);
		ok = ende_password_ok(ende.password, UD_DIGEST_SIZE);
	}

	crypt_size = ende_64_decode_size(coded_size);
	crypt_size = (crypt_size % 8) ? (crypt_size/8+1)*8 : crypt_size; // 8-byte boundary
	plain_size = crypt_size;
	total_size = title_size+plain_size;

	h_crypt = MemHandleNew(crypt_size);
	h_total = MemHandleNew(total_size);
	if (ok && h_crypt && h_total) {
		unsigned char *e = (unsigned char *) rec;	// encoded
		unsigned char *c = MemHandleLock(h_crypt);	// crypted
		unsigned char *p = MemHandleLock(h_total);	// plain
		if (c && p) {
			ULong i;
			MemSet(c, crypt_size, 0);
			MemSet(p, total_size, 0);
			MemMove(p, e, title_size);

			// Base64 Encode
			rc = ende_64_decode(e+title_size+ende.head_size, coded_size, c, crypt_size);

			if (ende.encrypt_list & UD_ENDE_DES & ende.encrypt_user) {

				// DES
				for (i = 0; i < crypt_size; i += 8) {

				 #ifdef UD_BLINK_WAIT_BITMAP
					if (i % 16) {
						uiDrawBitmap(EditWaitBitmap, UD_WAIT_BITMAP_X, UD_WAIT_BITMAP_Y);
					} else {
					 	uiEraseRect(UD_WAIT_BITMAP_X, UD_WAIT_BITMAP_Y, 10, 10);
					}
				 #endif
				 
				 #ifdef UD_1DES
					rc = EncDES(c+i,			ende.password,				  p+title_size+i, false);
				 #endif
				 
				 #ifdef UD_3DES
					rc = EncDES(c+i,			ende.password,				  p+title_size+i, false);
					rc = EncDES(p+title_size+i,	ende.password+UD_ENCKEY_SIZE, p+title_size+i, true);
					rc = EncDES(p+title_size+i,	ende.password,				  p+title_size+i, false);
				 #endif
				}
			} else {
				MemMove(p+title_size, c, crypt_size);
			}

			rc = MemHandleUnlock(h_total);
			rc = MemHandleUnlock(h_crypt);
		} else {
			ErrFatalDisplayIf(c == NULL, "DE Mem Error c");
			ErrFatalDisplayIf(p == NULL, "DE Mem Error p");
		}
		rc = MemHandleFree(h_crypt);
	} else {
		if (h_crypt) MemHandleFree(h_crypt);
		if (h_total) MemHandleFree(h_total);
		h_total = NULL;
	}

	return h_total;
}

void ende_encode_record(DmOpenRef db, Word record)
{
	VoidHand old_hand = NULL;
	VoidHand new_hand = NULL;
	ULong    old_size;			// original record size
	ULong    new_size;			// encoded record size
	ULong    title_size = 0;	// size of title (begin to 1st '\n' or eof)
	ULong    plain_size = 0;	// size of plain text body
	char*    old_ptr;
	char*    new_ptr;
	Err      rc;

	old_hand = DmQueryRecord(db, record);
	if (old_hand) {
		old_size = MemHandleSize(old_hand);
		old_ptr = MemHandleLock(old_hand);
		if (old_ptr) {
			// let first line as title
			char* nl = StrChr(old_ptr, '\n');
			if (nl) {
				title_size = nl - old_ptr + 1;
			} else {
				title_size = StrLen(old_ptr);
			}
			// plain text size, including tail '\0'
			plain_size = old_size - title_size;

			// encode text
			uiDrawBitmap(EditWaitBitmap, UD_WAIT_BITMAP_X, UD_WAIT_BITMAP_Y);
			new_hand = ende_encode_text(old_ptr, title_size, ende.head_size, plain_size);
		 	uiEraseRect(UD_WAIT_BITMAP_X, UD_WAIT_BITMAP_Y, 10, 10);

			rc = MemHandleUnlock(old_hand);
		}
	}

	if (new_hand) {
		new_size = MemHandleSize(new_hand);
		new_ptr = MemHandleLock(new_hand);
		old_hand = DmResizeRecord(db, record, new_size);
		if (old_hand) {
			old_ptr = MemHandleLock(old_hand);
		} else {
			old_ptr = NULL;
		}
		if (old_ptr && new_ptr) {
			rc = DmWrite(old_ptr, 0, new_ptr, new_size);
			rc = MemHandleUnlock(old_hand);
			rc = MemHandleUnlock(new_hand);
		} else {
			if (old_ptr) rc = MemHandleUnlock(old_hand);
			if (new_ptr) rc = MemHandleUnlock(new_hand);
		}
		if (old_hand) {
			rc = DmReleaseRecord(db, record, true);
		}
		rc = MemHandleFree(new_hand);
	}

 //	if (frm) FrmHideObject(frm, FrmGetObjectIndex(frm, EditWaitBitmap));
}

VoidHand ende_encode_text(char *rec, ULong title_size, ULong header_size, ULong plain_size)
{
	VoidHand h_crypt = NULL;
	VoidHand h_total = NULL;
	Boolean  ok = false;
	Err rc;

	ULong crypt_size, coded_size, total_size;

	crypt_size = (plain_size % 8) ? (plain_size/8+1)*8 : plain_size;
	coded_size = ende_64_encode_size(crypt_size);
	total_size = title_size+header_size+coded_size+sizeof('\0');

	if (ende_password_ok(ende.password, UD_DIGEST_SIZE)) {
		ok = true;
	} else {
		ende_password_set(ende.password);
		ok = ende_password_ok(ende.password, UD_DIGEST_SIZE);
	}

	if (ok) {
		h_crypt = MemHandleNew(crypt_size);
		h_total = MemHandleNew(total_size);
	}

	if (h_crypt && h_total) {
		unsigned char *p = (unsigned char *) rec;
		unsigned char *c = MemHandleLock(h_crypt);
		unsigned char *e = MemHandleLock(h_total);
		if (c && e) {
			ULong i;

			MemSet(c, crypt_size, 0);
			MemMove(c, p+title_size, plain_size);

			if (ende.encrypt_list & UD_ENDE_DES & ende.encrypt_user) {
				for (i = 0; i < crypt_size; i += 8) {

				 #ifdef UD_BLINK_WAIT_BITMAP
					if (i % 16) {
						uiDrawBitmap(EditWaitBitmap, UD_WAIT_BITMAP_X, UD_WAIT_BITMAP_Y);
					} else {
					 	uiEraseRect(UD_WAIT_BITMAP_X, UD_WAIT_BITMAP_Y, 10, 10);
					}
				 #endif

				 #ifdef UD_1DES
					rc = EncDES(c+i, ende.password,                c+i, true);
				 #endif
				 
				 #ifdef UD_3DES
					rc = EncDES(c+i, ende.password,                c+i, true);
					rc = EncDES(c+i, ende.password+UD_ENCKEY_SIZE, c+i, false);
					rc = EncDES(c+i, ende.password,                c+i, true);
				 #endif
				}
			}

			// Base64 Encode
			MemSet(e, total_size, 0);
			MemMove(e, p, title_size);
			MemMove(e+title_size, ende.signature, ende.sign_size);
			MemMove(e+title_size+ende.sign_size, ende.checksum, ende.csum_size);
			rc = ende_64_encode(c, crypt_size, e+title_size+ende.head_size, coded_size);
			{
			char *cs = (char*) e+title_size+ende.sign_size;
			ende_set_checksum(cs, Crc16CalcBlock(e+title_size+ende.head_size, coded_size, 0));
			}
			rc = MemHandleUnlock(h_total);
			rc = MemHandleUnlock(h_crypt);
		} else {
			ErrFatalDisplayIf(c == NULL, "DE Mem Error");
			ErrFatalDisplayIf(e == NULL, "DE Mem Error");
		}
		rc = MemHandleFree(h_crypt);
	} else {
		if (h_crypt) MemHandleFree(h_crypt);
		if (h_total) MemHandleFree(h_total);
		h_total = NULL;
	}

	return h_total;
}

ULong ende_64_encode_size(ULong de_size)
{
	ULong encode_size = (de_size % 3) ? (de_size / 3 + 1) * 4 : de_size * 4 / 3;	// 4 bytes to store 3 bytes
	return encode_size;
}

ULong ende_64_decode_size(ULong en_size)
{
	ULong decode_size = (en_size % 4) ? (en_size / 3 + 1) * 4 : en_size * 3 / 4; 	// 3 bytes to store 4 bytes
	return decode_size;
}

Err ende_64_encode(unsigned char *de, Long de_size, unsigned char *en, Long en_size)
{
	Long de_i = 0, en_i = 0;
	for (de_i = 0; de_i <= de_size-3; de_i += 3, en_i += 4) {
		*en++ = BASE64[de[de_i] >> 2];
		*en++ = BASE64[((de[de_i]   & 0x03) << 4) | (de[de_i+1] >> 4)];
		*en++ = BASE64[((de[de_i+1] & 0x0f) << 2) | (de[de_i+2] >> 6)];
		*en++ = BASE64[de[de_i+2] & 0x3f];
	}

	if (de_i != de_size) {
		if (de_size - de_i == 2) {
			*en++ = BASE64[de[de_i] >> 2];
			*en++ = BASE64[((de[de_i]   & 0x03) << 4) | ((de[de_i+1] & 0xf0) >> 4)];
			*en++ = BASE64[((de[de_i+1] & 0x0f) << 2)];
			*en++ = '=';
		} else if (de_size - de_i == 1) {
			*en++ = BASE64[de[de_i] >> 2];
			*en++ = BASE64[(de[de_i] & 0x03) << 4];
			*en++ = '=';
			*en++ = '=';
		}
		en_i += 4;
    }

	ErrFatalDisplayIf(en_i != en_size, "B64EN");

	return 0;
}

Err ende_64_decode(unsigned char *en, Long en_size, unsigned char *de, Long de_size)
{
	Long en_i = 0, de_i = 0;

	while (en_i < en_size && de_i < de_size) {
		if (en_i+1 < en_size) {
			*de++ = ((ASCII[(en[en_i]   & 0x7f)] & 0x3f) << 2) | ((ASCII[(en[en_i+1] & 0x7f)] & 0x30) >> 4);
		}
		if (en_i+2 < en_size) {
			*de++ = ((ASCII[(en[en_i+1] & 0x7f)] & 0x0f) << 4) | ((ASCII[(en[en_i+2] & 0x7f)] & 0x3c) >> 2);
		}
		if (en_i+3 < en_size) {
			*de++ = ((ASCII[(en[en_i+2] & 0x7f)] & 0x03) << 6) | ((ASCII[(en[en_i+3] & 0x7f)] & 0x3f));
		}
		en_i += 4;
		de_i += 3;
	}

//	ErrFatalDisplayIf((en_size%4)!=0, "B64DE Alignment Error");

	return 0;
}

Boolean ende_digest_defed(Byte digest[16], Byte size)
{
	Byte i;
	for (i = 0; i < size; i++) {
		if (digest[i] != 0) return true;
	}
	return false;
}

Boolean ende_password_1st()
{
	Word rc = 0;
	Byte i, match;
	Byte pw1st[UD_DIGEST_SIZE];
	Byte pw2nd[UD_DIGEST_SIZE];
	Boolean again = false;
	Boolean ok = false;
	
	MemSet(&pw1st, UD_DIGEST_SIZE, 0x01);
	MemSet(&pw2nd, UD_DIGEST_SIZE, 0x02);
	
	do {
		FrmCustomAlert(AnyAlert, 
			"Please select a private password to protect your data."
			" You will enter the password twice to verify.\n"
			"Please remember the password carefully."
			" If you forget it, there is no way to recover any encrypted data.",
			" ", 
			" ");

		if (ende_ask_password("New Password", pw1st) && 
			ende_ask_password("New Password Again", pw2nd))
		{
			for (i = 0, match = 0; i < UD_DIGEST_SIZE; i++) {
				if (pw1st[i] == pw2nd[i]) match++;
			}

			if (match == UD_DIGEST_SIZE) {
				rc = EncDigestMD5(pw1st, sizeof(pw1st), pref.pwdod);	// store digest of password
				MemMove(&ende.password, &pw1st, sizeof(ende.password));	// store run-time password
				ok = true;
			} else {
			 //	MemSet(pref.pwdod, UD_DIGEST_SIZE, 0);
			 //	MemSet(ende.password, UD_DIGEST_SIZE, 0);
			}
		}

		again = false;

		if (!ok) {
			rc = FrmCustomAlert(AskAlert, "Password is not properly defined!\n\nTry Again?", " ", " ");
			if (rc == 0) again = true;
		}
	} while (again);
	
	MemSet(&pw1st, UD_DIGEST_SIZE, 0);
	MemSet(&pw2nd, UD_DIGEST_SIZE, 0);
	
	return ok;
}

void ende_password_set(Byte pw[UD_DIGEST_SIZE])
{
	Word rc = 0;
	Byte i, match;
	Byte digest[UD_DIGEST_SIZE];
	Boolean pw_ok = false;
	Boolean again = false;

	do {
		if (ende_ask_password("Enter Password", pw)) {
			;
		}
		rc = EncDigestMD5(pw, UD_DIGEST_SIZE, digest);
		for (i = 0, match = 0; i < UD_DIGEST_SIZE; i++) {
			if (digest[i] == pref.pwdod[i]) match++;
		}
		if (match == UD_DIGEST_SIZE) {
			pw_ok = true;
			again = false;
		} else {
			pw_ok = false;
			again = false;
			MemSet(pw, UD_DIGEST_SIZE, 0);
			rc = FrmCustomAlert(AskAlert, "Password incorrect!\nTry Again?", " ", " ");
			if (rc == 0) again = true;
		}

	} while (again);
}

Boolean ende_password_ok(Byte pw[UD_DIGEST_SIZE], Byte size)
{
	Boolean ok = false;
	Byte digest[UD_DIGEST_SIZE];

	if (EncDigestMD5((Byte*) pw, size, digest) == UD_NO_ERROR) {
		Byte i, match;
		for (i = 0, match = 0; i < UD_DIGEST_SIZE; i++) {
			if (digest[i] == pref.pwdod[i]) match++;
		}
		if (match == UD_DIGEST_SIZE) {
			ok = true;
		} else {
			MemSet(ende.password, UD_DIGEST_SIZE, 0);
		}
	}

	return ok;
}

Boolean	ende_ask_password(char *title, Byte pw[16])
{
	Boolean ok = false;

	FormPtr prevFormP = FrmGetActiveForm();
	FormPtr currFormP = FrmInitForm(PassForm);
	if (currFormP) {
		if (StrLen(title) <= 30) FrmCopyTitle(currFormP, title);
		FrmSetFocus(currFormP, FrmGetObjectIndex(currFormP, PassPasswordField));
		FrmSetActiveForm(currFormP);
		if (FrmDoDialog(currFormP) == PassOKButton) {
			char *pp = uiGetText(currFormP, PassPasswordField);
			UInt len = StrLen(pp);
			if (len > 0 && EncDigestMD5((Byte*) pp, len, pw) == UD_NO_ERROR) {
				ok = true;		
			}
		}
		FrmDeleteForm(currFormP);
	}
	if (prevFormP) FrmSetActiveForm(prevFormP);

	return ok;
}

void ende_change_password()
{
	Boolean pw_changed = false;
	Word rc = FrmCustomAlert(AskAlert,
		"If you change to a new password, any memo encrypted with the old password will not be readable.\n\n"
		"Are you sure to change to a new password?\n",
		" ", 
		" ");
	if (rc == 0) {
		Byte pw[UD_DIGEST_SIZE];
		if (ende_ask_password("Current Password", pw) && ende_password_ok(pw, UD_DIGEST_SIZE)) {
			if (ende_password_1st()) {
				pw_changed = true;
			} else {
				;
			}
		} else {
			;
		}
	}
	
	if (!pw_changed) FrmCustomAlert(AnyAlert, "Current password is not changed.", " ", " ");
}

void ende_forget_password()
{
	MemSet(ende.password, sizeof(ende.password), 0);
}

void uiDrawBitmap(int bmID, short scrnx, short scrny)
{
	if (bmID) { // a bitmap selected
		VoidHand handle = DmGetResource(bitmapRsc, bmID);
		if (handle) {
			void* memory = MemHandleLock(handle);
			if (memory) {
				// draw the piece
				WinDrawBitmap(memory, scrnx, scrny);
				MemPtrUnlock(memory);
			}
			DmReleaseResource(handle);
		}
	}
}

void uiEraseRect(short scrnx, short scrny, short bmW, short bmH)
{
	RectangleType r;
	RctSetRectangle(&r, scrnx, scrny, bmW, bmH);
	WinEraseRectangle(&r, 0);
}

void uiSetText(FormPtr frm, Word fID, CharPtr fData, ULong fSize)
{
	Word fldIndex = FrmGetObjectIndex(frm, fID);
	FieldPtr fld = (FieldPtr) FrmGetObjectPtr(frm, fldIndex);
	VoidHand fieldHandle = MemHandleNew(fSize);
	VoidPtr  fieldData = MemHandleLock(fieldHandle);
	StrCopy((CharPtr) fieldData, fData);
	MemHandleUnlock(fieldHandle);
	FldSetTextHandle(fld, (Handle) fieldHandle);
}

char *uiGetText(FormPtr frm, Word fID)
{
	Word fldIndex = FrmGetObjectIndex(frm, fID);
	FieldPtr fld = (FieldPtr) FrmGetObjectPtr(frm, fldIndex);
	CharPtr fieldData = FldGetTextPtr(fld);
	return fieldData ? fieldData : "";
}

void uiAboutSafe()
{
	FormPtr prevFormP = FrmGetActiveForm();
	FormPtr currFormP = FrmInitForm(SafeDialog);
	FrmSetActiveForm(currFormP);
  #ifdef UD_UNREGISTER 
	FrmShowObject(currFormP, FrmGetObjectIndex(currFormP, SafeUnRegLabel));
  #endif
	FrmDoDialog(currFormP);
	FrmDeleteForm(currFormP);
	if (prevFormP) FrmSetActiveForm(prevFormP);
}

